install_deps = ['numpy>=1.22.4', 'scipy', 'numba', 
                'edt','scikit-image','ncolor>=1.2.1',
                'scikit-learn','torch>=1.10',
                'mahotas>=1.4.13',
                'mgen','matplotlib',
                'peakdetect','igraph',
                'torchvf','tqdm',
                ]

gui_deps = [
        'pyqtgraph>=0.12.4', 
        'PyQt6.sip', 
        'PyQt6',
        'google-cloud-storage',
        'omnipose-theme',
        'superqt','darkdetect'
        ]

doc_deps = ['sphinx-autobuild',
            'sphinx_automodapi',
            'sphinx_copybutton',
            'sphinx_design','furo','myst_nb']

distributed_deps = [
        'dask',
        'dask_image',
        'scikit-learn',
        ]